package com.reactlibrary

import android.app.Activity
import android.content.Intent
import com.facebook.react.bridge.*
import com.facebook.react.modules.core.RCTNativeAppEventEmitter
import com.google.android.gms.common.api.ApiException
import com.google.android.gms.common.api.CommonStatusCodes
import com.google.android.gms.common.api.ResolvableApiException
import com.google.android.gms.wallet.PaymentData
import com.google.android.gms.wallet.WalletConstants
import io.inai.android_sdk.*
import io.inai.android_sdk.InaiCheckout
import kotlinx.serialization.decodeFromString
import org.json.JSONObject
import kotlinx.serialization.json.Json
import org.json.JSONArray

private const val PAYMENT_SUCCESS = "Inai::PAYMENT_SUCCESS"
private const val PAYMENT_ERROR = "Inai::PAYMENT_ERROR"

class InaiCheckout(reactContext: ReactApplicationContext) : ReactContextBaseJavaModule(reactContext),
    InaiCheckoutDelegate, InaiCardInfoDelegate, InaiValidateFieldsDelegate, InaiPaymentMethodDelegate, InaiInstallmentPlansDelegate, ActivityEventListener {

    private var googlePayRequestData: InaiGooglePayRequestData? = null
    private var config: InaiConfig? = null
    private var GOOGLE_PAY = "google_pay"
    private var reactContext: ReactApplicationContext

    init {
        this.reactContext = reactContext
        reactContext.addActivityEventListener(this)
    }

    override fun getName(): String {
        return "RNInaiSdk"
    }


    private fun getInaiConfigObjectFromMap(map: ReadableMap): InaiConfig {
        val jsonString = JSONObject(map.toHashMap()).toString()
        return Json.decodeFromString(jsonString)
    }

    @ReactMethod
    fun addListener(eventName: String) {
        // Set up any upstream listeners or background tasks as necessary
    }

    @ReactMethod
    fun removeListeners(count: Int) {
        // Remove upstream listeners, stop unnecessary background tasks
    }

    @ReactMethod
    fun presentCheckout(
        config: ReadableMap
    ) {
        try {
            val inaiConfig = getInaiConfigObjectFromMap(config)
            val checkout = InaiCheckout(inaiConfig)
            currentActivity.let {
                if (it != null) {
                    checkout.presentCheckout(it, this)
                }
            }
        } catch (e: Exception) {
            val wMap = WritableNativeMap()
            wMap.putString("Init Failed", e.message)
            sendEvent(PAYMENT_ERROR, wMap)
        }
    }

    @ReactMethod
    fun makePayment(
        config: ReadableMap,
        paymentMethodOption: String,
        paymentDetails: ReadableMap
    ) {
        try {
            val inaiConfig = getInaiConfigObjectFromMap(config)
            val checkout = InaiCheckout(inaiConfig)
            val paymentDetailsJSON = JSONObject(paymentDetails.toHashMap())
            currentActivity?.let {
                if (it != null) {
                    checkout.makePayment(paymentMethodOption, paymentDetailsJSON, it, this)
                }
            }
        } catch (e: Exception) {
            val wMap = WritableNativeMap()
            wMap.putString("Init Failed", e.message)
            sendEvent(PAYMENT_ERROR, wMap)
        }
    }

    @ReactMethod
    fun getInstallmentPlans(
        config: ReadableMap,
        paymentMethodOption: String, 
        paymentDetails: ReadableMap
    ) {
        try {
            val inaiConfig = getInaiConfigObjectFromMap(config)
            val checkout = InaiCheckout(inaiConfig)
            val paymentDetailsJSON = JSONObject(paymentDetails.toHashMap())
            currentActivity.let {
                if (it != null) {
                    checkout.getInstallmentPlans(paymentMethodOption, paymentDetailsJSON, it, this)
                }
            }
        } catch (e: Exception) {
            val wMap = WritableNativeMap()
            wMap.putString("Init Failed", e.message)
            sendEvent(PAYMENT_ERROR, wMap)
        }
    }

    @ReactMethod
    fun addPaymentMethod(
        config: ReadableMap,
        paymentMethodType: String
    ) {
        try {
            val inaiConfig = getInaiConfigObjectFromMap(config)
            val checkout = InaiCheckout(inaiConfig)
            currentActivity?.let {
                checkout.addPaymentMethod(paymentMethodType, it, this)
            }
        } catch (e: Exception) {
            val wMap = WritableNativeMap()
            wMap.putString("Init Failed", e.message)
            sendEvent(PAYMENT_ERROR, wMap)
        }
    }

    @ReactMethod
    fun payWithPaymentMethod(
        config: ReadableMap,
        paymentMethodId: String
    ){
        try {
            val inaiConfig = getInaiConfigObjectFromMap(config)
            val checkout = InaiCheckout(inaiConfig)
            currentActivity?.let {
                checkout.presentPayWithPaymentMethod(paymentMethodId,it, this)
            }
        } catch (e: Exception) {
            val wMap = WritableNativeMap()
            wMap.putString("Init Failed", e.message)
            sendEvent(PAYMENT_ERROR, wMap)
        }
    }

    @ReactMethod
    fun validateFields(
        config: ReadableMap,
        paymentMethodOption: String,
        paymentDetails: ReadableMap
    ) {
        try {
            val inaiConfig = getInaiConfigObjectFromMap(config)
            val checkout = InaiCheckout(inaiConfig)
            val paymentDetailsJSON = JSONObject(paymentDetails.toHashMap())
            currentActivity.let {
                if (it != null) {
                    checkout.validateFields(paymentMethodOption, paymentDetailsJSON, it, this)
                }
            }
        } catch (e: Exception) {
            val wMap = WritableNativeMap()
            wMap.putString("Init Failed", e.message)
            sendEvent(PAYMENT_ERROR, wMap)
        }
    }

    @ReactMethod
    fun getCardInfo(
        config: ReadableMap,
        cardNumber: String
    ) {
        try {
            val inaiConfig = getInaiConfigObjectFromMap(config)
            val checkout = InaiCheckout(inaiConfig)
            currentActivity.let {
                if (it != null) {
                    checkout.getCardInfo(cardNumber, it, this)
                }
            }
        } catch (e: Exception) {
            val wMap = WritableNativeMap()
            wMap.putString("Init Failed", e.message)
            sendEvent(PAYMENT_ERROR, wMap)
        }
    }

    @ReactMethod
    fun initGooglePay(
        paymentDetails: ReadableMap
    ) {
        val paymentDetailsJSON = JSONObject(paymentDetails.toHashMap())
        currentActivity.let {
            if (it != null) {
                try {
                    InaiCheckout.initGooglePay(
                        paymentDetailsJSON,
                        WalletConstants.ENVIRONMENT_TEST,
                        it
                    ) { googlePayRequestData ->
                        if (googlePayRequestData != null) {
                            val wMap = WritableNativeMap()
                            if (googlePayRequestData.canMakePayments) {
                                // Send back data
                                this.googlePayRequestData = googlePayRequestData
                                wMap.putString("${googlePayRequestData.canMakePayments}", null)
                                sendEvent(PAYMENT_SUCCESS, wMap)
                            } else {
                                wMap.putString("${googlePayRequestData.canMakePayments}", null)
                                sendEvent(PAYMENT_SUCCESS, wMap)
                            }
                        } else {
                            val wMap = WritableNativeMap()
                            wMap.putString("Error", "Google Pay Not Available")
                            sendEvent(PAYMENT_ERROR, wMap)
                        }
                    }
                } catch (e: Exception) {
                    val wMap = WritableNativeMap()
                    wMap.putString("Init Failed", e.message)
                    sendEvent(PAYMENT_ERROR, wMap)
                }
            }
        }
    }

    @ReactMethod
    fun launchGooglePay(
        config: ReadableMap
    ) {
        this.config = getInaiConfigObjectFromMap(config)
        googlePayRequestData?.let {
            val task = InaiCheckout.launchGooglePayRequest(googlePayRequestData!!)
            task.addOnCompleteListener { completedTask ->
                if (completedTask.isSuccessful) {
                    completedTask.result?.let { paymentData ->
                        val paymentDataJsonString = paymentData.toJson()
                        this.config?.let {
                            handleGooglePaySuccess(paymentDataJsonString, GOOGLE_PAY, it)
                        }
                    }
                } else {
                    when (val exception = completedTask.exception) {
                        is ResolvableApiException -> {
                            currentActivity?.let { exception.startResolutionForResult(it, 101) }
                        }
                        is ApiException -> {
                            val errorStr =
                                "ApiException Error code ${exception.statusCode} ${exception.message}"
                            val wMap = WritableNativeMap()
                            wMap.putString("Error", errorStr)
                            sendEvent(PAYMENT_ERROR, wMap)
                        }
                        else -> {
                            val errorStr =
                                "Error ${CommonStatusCodes.INTERNAL_ERROR} Unexpected exception"
                            val wMap = WritableNativeMap()
                            wMap.putString("Init Failed", errorStr)
                            sendEvent(PAYMENT_ERROR, wMap)
                        }
                    }
                }
            }
        }
    }


    private fun handleGooglePaySuccess(
        paymentDataJSON: String,
        paymentMethodOption: String,
        config: InaiConfig,

        ) {
        //  Google Auth successful
        //  Init Inai SDK and process checkout
        try {
            //  Process google token for payment details data
            val paymentData = PaymentData.fromJson(paymentDataJSON)
            val paymentDetails = InaiCheckout.getGooglePayRequestData(paymentData)
            val checkout = InaiCheckout(config)
            currentActivity?.let {
                checkout.makePayment(paymentMethodOption, paymentDetails, it, this)
            }

        } catch (ex: Exception) {
            //  Handle initialization error
            val wMap = WritableNativeMap()
            wMap.putString("Init Failed", ex.message)
            sendEvent(PAYMENT_ERROR, wMap)
        }
    }

    override fun paymentFinished(result: InaiPaymentResult) {
        //  Put only `data` from result into a JSON Object.
        val resultData: JSONObject = result.data

        if (result.status == InaiPaymentStatus.Success) {
            resolvePromise(resultData)
        } else {
            rejectPromise(resultData)
        }
    }

    override fun cardInfoFetched(result: InaiCardInfoResult) {
        //  Put only `data` from result into a JSON Object.
        val resultData: JSONObject = result.data

        if (result.status == InaiCardInfoStatus.Success) {
            resolvePromise(resultData)
        } else {
            rejectPromise(resultData)
        }
    }

    override fun fieldsValidationFinished(result: InaiValidateFieldsResult) {
        //  Put only `data` from result into a JSON Object.
        val resultData: JSONObject = result.data

        if (result.status == InaiValidateFieldsStatus.Success) {
            resolvePromise(resultData)
        } else {
            rejectPromise(resultData)
        }
    }

    override fun installmentPlansFetched(result: InaiInstallmentPlansResult) {
        val resultData: JSONObject = result.data
        if (result.status == InaiInstallmentPlansInfoStatus.Success) {
            resolvePromise(resultData)
        } else {
            rejectPromise(resultData)
        }
    }

    override fun paymentMethodSaved(result: InaiPaymentMethodResult) {
        //  Put only `data` from result into a JSON Object.
        val resultData: JSONObject = result.data

        if (result.status == InaiPaymentMethodStatus.Success) {
            resolvePromise(resultData)
        } else {
            rejectPromise(resultData)
        }
    }

    override fun onActivityResult(activity: Activity?, requestCode: Int, resultCode: Int, intentData: Intent?) {
        when (resultCode) {
            Activity.RESULT_OK ->
                intentData?.let { intent ->
                    PaymentData.getFromIntent(intent)?.let { paymentData ->
                        val paymentDataJsonString = paymentData.toJson()
                        config?.let {
                            handleGooglePaySuccess(paymentDataJsonString, GOOGLE_PAY, it)
                        }
                    }
                }

            Activity.RESULT_CANCELED -> {
                // The user canceled the payment attempt
                val wMap = WritableNativeMap()
                wMap.putString("Error", "Google Pay cancelled")
                sendEvent(PAYMENT_ERROR, wMap)
            }
        }
    }

    override fun onNewIntent(intent: Intent?) {

    }

    private fun resolvePromise(resultData: JSONObject) {
        //  Convert JSON Object to WritableMap
        val writableMap = convertJSONToWritableMap(resultData)
        //  Resolve Promise
        sendEvent(PAYMENT_SUCCESS, writableMap);
    }

    private fun rejectPromise(resultData: JSONObject) {
        //  Convert JSON Object to WritableMap
        val writableMap = convertJSONToWritableMap(resultData)
        //  Reject Promise
        sendEvent(PAYMENT_ERROR, writableMap);
    }

    private fun sendEvent(eventName: String, params: WritableMap) {
        reactContext
            .getJSModule(RCTNativeAppEventEmitter::class.java)
            .emit(eventName, params)
    }

    private fun convertJSONToWritableMap(jsonObject: JSONObject): WritableMap {
        val wMap = WritableNativeMap()
        val keys = jsonObject.keys()
        while (keys.hasNext()) {
            val key = keys.next()
            val value = jsonObject.get(key)
            if (value == JSONObject.NULL) {
                wMap.putNull(key)
            }
            when (value) {
                is String -> {
                    wMap.putString(key, value)
                }
                is Int -> {
                    wMap.putInt(key, value)
                }
                is Boolean -> {
                    wMap.putBoolean(key, value)
                }
                is JSONObject -> {
                    val map = convertJSONToWritableMap(value)
                    wMap.putMap(key, map)
                }
                is Double -> {
                    wMap.putDouble(key, value) 
                }
                is JSONArray -> {
                    val maps = WritableNativeArray()
                    for (i in 0 until value.length()) {
                        maps.pushMap(convertJSONToWritableMap(value.getJSONObject(i)))
                    }
                    wMap.putArray(key, maps)
                }
            }
        }
        return wMap
    } 
}